import math
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from .img_types import Img

def show(
  *inputs: Img | tuple[str, Img],
  ncols: int = 3, height: float = 5, width: float = 5,
  axis = False, cmap = "gray"
) -> Figure:
  """Shows a set of images or (title, image) pairs"""

  def show_input(ax: Axes, input: Img | tuple[str, Img]):
    match input:
      case title, img:
        ax.imshow(img, cmap=cmap)
        ax.set_title(title)
      case img:
        ax.imshow(img, cmap=cmap)
    ax.axis(axis)
    
  n = len(inputs)
  if n == 0:
    return Figure()
  if n == 1:
    fig, ax = plt.subplots(figsize=(width, height))
    show_input(ax, inputs[0])
  elif n <= ncols:
    fig, ax = plt.subplots(1, n, figsize=(5*n, height))
    for i, input in enumerate(inputs):
      show_input(ax[i], input)
  elif ncols == 1:
    nrows = n
    fig, ax = plt.subplots(nrows, 1, figsize=(5, height*nrows))
    for i, input in enumerate(inputs):
      show_input(ax[i], input)
  else:
    nrows = math.ceil(n / ncols)
    fig, ax = plt.subplots(nrows, ncols, figsize=(5*ncols, height*nrows))
    for i, input in enumerate(inputs):
      show_input(ax[i // ncols, i % ncols], input)
  
  plt.close(fig)
  return fig


import colorsys

def mod_color(i: int, n: int, lightness: float = 0.5, saturation: float = 0.8) -> tuple[int, int, int]:
    """Color per every `i mod n`, evenly spaced around the hue circle"""
    hue = i / n
    rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
    return tuple(int(x*255) for x in rgb)


from typing import Sequence, overload, Callable
from jaxtyping import Shaped
import cv2 as cv
import numpy as np
import ramda as R
from ..img_types import Img

Vec2 = Shaped[np.ndarray, '2']
Vertices = Shaped[np.ndarray, 'N 2'] | Sequence[Vec2]

@overload
def vertices(img: Img, /, vertices: Vertices, *, radius: int = 10, color: int | tuple[int, int, int] = (255, 0, 0)) -> Img: ...
@overload
def vertices(*, vertices: Vertices, radius: int = 10, color: int | tuple[int, int, int] = (255, 0, 0)) -> Callable[[Img], Img]: ...

@R.curry
def vertices(
    img: Img, /, vertices: Vertices, radius: int = 10, color: int | tuple[int, int, int] = (255, 0, 0)
) -> Img: # type: ignore
    if isinstance(color, tuple):
        img = cv.cvtColor(img, cv.COLOR_GRAY2BGR) if len(np.shape(img)) == 2 else np.copy(img)
    output = img.copy()
    for [x, y] in vertices:
        cv.circle(output, (int(x), int(y)), radius=radius, color=color, thickness=-1)
    return output


@R.curry
def vlines(xs: list[float], img: Img, color = (255, 0, 0), thickness = 3):
  height = img.shape[0]
  ls = np.array([[[x, 0, x, height-1]] for x in xs], dtype=np.int32)
  return lines(img, ls, color=color, thickness=thickness)

@R.curry
def hlines(ys: list[float], img: Img, color = (255, 0, 0), thickness = 3):
  width = img.shape[1]
  ls = np.array([[[0, y, width-1, y]] for y in ys], dtype=np.int32)
  return lines(img, ls, color=color, thickness=thickness)

Segment = Shaped[np.ndarray, '1 4']
Segments = Shaped[np.ndarray, 'N 1 4'] | Sequence[Segment]

@overload
def lines(img: Img, /, lines: Segments, *, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Img: ...
@overload
def lines(*, lines: Segments, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Callable[[Img], Img]: ...
@R.curry
def lines(
  img: Img, /, lines: Segments, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3
) -> Img: # type: ignore
  out = img.copy()
  for line in lines:
    x1, y1, x2, y2 = line[0]
    cv.line(out, [x1, y1], [x2, y2], color=color, thickness=thickness)
  return out


@overload
def contours(img: Img, /, contours: Sequence[Shaped[np.ndarray, 'N 2']], *, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Img: ...
@overload
def contours(*, contours: Sequence[Shaped[np.ndarray, 'N 2']], color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Callable[[Img], Img]: ...
@R.curry
def contours(
  img: Img, /, contours: Sequence[Shaped[np.ndarray, 'N 2']], color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3
) -> Img: # type: ignore
  if isinstance(color, tuple):
    outimg = cv.cvtColor(img, cv.COLOR_GRAY2BGR) if len(np.shape(img)) == 2 else np.copy(img)
  else:
    outimg = np.copy(img) if len(np.shape(img)) == 2 else cv.cvtColor(img, cv.COLOR_BGR2GRAY)
    
  return cv.drawContours(outimg, [np.array(c, dtype=np.int32) for c in contours], -1, color, thickness)




from typing import TypeAlias
from jaxtyping import Shaped
import numpy as np

Img2D: TypeAlias = Shaped[np.ndarray, "w h"]
Img3D: TypeAlias = Shaped[np.ndarray, "w h c"]
Img: TypeAlias = Img2D | Img3D | np.ndarray


from typing import Callable, overload
import cv2 as cv
import ramda as R
from .img_types import Img

@overload
def rescale_h(img: Img, /, height: int) -> Img: ...
@overload
def rescale_h(*, height: int) -> Callable[[Img], Img]: ...

@R.curry
def rescale_h(img: Img, /, height: int) -> Img: # type: ignore
    """Rescales the image to `height`"""
    h, w = img.shape[:2]
    width = int(height * w / h)
    return cv.resize(img, (width, height))

@overload
def descale_h(img: Img, /, target_height: int) -> Img: ...
@overload
def descale_h(*, target_height: int) -> Callable[[Img], Img]: ...

@R.curry
def descale_h(img: Img, /, target_height: int) -> Img: # type: ignore
    """Rescales the image to `height = min(img.height, target_height)`"""
    cur_height, _ = img.shape[:2]
    return rescale_h(img, target_height) if target_height < cur_height else img



from typing import Literal
import cv2 as cv
from .types_ import Rotation
from .img_types import Img

def code(rot: Rotation) -> int:
  if rot == 90:
    return cv.ROTATE_90_COUNTERCLOCKWISE
  if rot == -90:
    return cv.ROTATE_90_CLOCKWISE
  if rot == 180:
    return cv.ROTATE_180
  else:
    raise ValueError(f'Invalid rotation {rot}')
  
def rotate(img: Img, rotation: Rotation | Literal[0]) -> Img:
  if rotation == 0:
    return img
  return cv.rotate(img, code(rotation))


from .types_ import Rotation
from .img_types import Img, Img2D, Img3D
from .opencv import imread, grayscale, threshold, dilate, erode, cvtColor, imdecode
try:
  from .plot import show
except ImportError:
  ...
from .encoding import b64encode, b64decode, encode, decode
from .colors import mod_color
from .misc import black
from .rescaling import descale_h, rescale_h
from .rotation import rotate
from . import draw

__all__ = [
  'imread', 'grayscale', 'threshold', 'dilate', 'erode', 'cvtColor', 'imdecode', 
  'show',
  'b64encode', 'b64decode', 'encode', 'decode',
  'mod_color',
  'black',
  'descale_h', 'rescale_h',
  'rotate', 'Rotation',
  'draw',
  'Img', 'Img2D', 'Img3D',
]



from typing import Literal, overload, Callable
import base64
import numpy as np
import cv2 as cv
import numpy as np
import ramda as R
from .img_types import Img

@overload
def encode(img: Img, format: Literal[".png", ".jpg", ".webp"]) -> bytes: ...
@overload
def encode(*, format: Literal[".png", ".jpg", ".webp"]) -> Callable[[Img], bytes]: ...

@R.curry
def encode(img: Img, format: Literal[".png", ".jpg", ".webp"]) -> bytes: # type: ignore
    _, encoded = cv.imencode(format, img)
    return encoded.tobytes()

@overload
def b64encode(img: Img, format: Literal[".png", ".jpg", ".webp"] = ".png", url_safe: bool = True) -> str: ...
@overload
def b64encode(*, format: Literal[".png", ".jpg", ".webp"] = ".png", url_safe: bool = True) -> Callable[[Img], str]: ...

@R.curry
def b64encode(img: Img, format: Literal[".png", ".jpg", ".webp"] = ".png", url_safe: bool = True) -> str: # type: ignore
    """Encode `img` into a Base64 string"""
    _, enc = cv.imencode(format, img)
    encoded = bytes(enc)
    b64 = base64.urlsafe_b64encode(encoded) if url_safe else base64.b64encode(encoded)
    return b64.decode("utf-8")

def decode(img: bytes, flags = cv.IMREAD_UNCHANGED) -> Img:
    arr = np.frombuffer(img, np.uint8)
    return cv.imdecode(arr, flags)

def b64decode(b64img: str) -> Img:
    b = base64.urlsafe_b64decode(b64img)
    arr = np.frombuffer(b, np.uint8)
    return cv.imdecode(arr, -1)



from typing import overload, Callable
import cv2 as cv
import numpy as np
import ramda as R
from .img_types import Img2D, Img

Mat3x3 = cv.Mat
_N = int
_2 = int

@R.curry
def imdecode(img, flags = cv.IMREAD_UNCHANGED) -> Img:
  return cv.imdecode(img, flags)

def imread(path: str) -> Img:
  """Read in RGB (instead of BGR)"""
  img = cv.imread(path)
  return cv.cvtColor(img, cv.COLOR_BGR2RGB)

def grayscale(img: Img) -> Img2D:
  """`Array[h, w, ...] -> Array[h, w]`"""
  return img if len(np.shape(img)) == 2 else cv.cvtColor(img, cv.COLOR_BGR2GRAY)

def threshold(img: Img) -> Img2D:
  """`Array[h, w, ...] -> Array[h, w]` with values in `{0, 255}` only"""
  gray_img = grayscale(img)
  return cv.adaptiveThreshold(~gray_img, 255, cv.ADAPTIVE_THRESH_MEAN_C, cv.THRESH_BINARY, 15, -3)


@overload
def cvtColor(src: Img, code: int) -> Img: ...
@overload
def cvtColor(*, code: int) -> Callable[[Img], Img]: ...
@R.curry
def cvtColor(src: Img, /, code: int) -> Img: # type: ignore
  return cv.cvtColor(src, code)

@overload
def dilate(src: Img, /, kernel: cv.Mat, iterations: int) -> Img: ...
@overload
def dilate(*, kernel: cv.Mat, iterations: int) -> Callable[[Img], Img]: ...

@R.curry
def dilate(src: Img, /, kernel: cv.Mat, iterations: int) -> Img: # type: ignore
    return cv.dilate(src, kernel=kernel, iterations=iterations)

@overload
def erode(src: Img, /, kernel: cv.Mat, iterations: int) -> Img: ...
@overload
def erode(*, kernel: cv.Mat, iterations: int) -> Callable[[Img], Img]: ...

@R.curry
def erode(src: Img, /, kernel: cv.Mat, iterations: int) -> Img: # type: ignore
    return cv.erode(src, kernel=kernel, iterations=iterations)